#include "cfgs.h"
#include "fsio.h"
#include "centroid.h"
#include "binaryinfo.h"

#include <string.h>

//time measurement
#include <tchar.h>
#include <windows.h>

//directory used for temporary files and error logging
static const char *centroid_analysis_tmpdir = "D:\\centroid_analysis_tmp";

//globally visible data regarding error logging file 
char errlog_file[MAX_PATH];
char errlog_entry[2 * MAX_PATH];

//connection strings for the mongo masters and slaves (master is renegotiated after each reboot of a machine) 
static const char* mongo_connstrings[4] = { "mongodb://localhost:27017", "mongodb://localhost:27018", "mongodb://localhost:27019", "mongodb://localhost:27020" };

#if defined(_WIN32)
#include <winsock2.h>
#endif
#include "mongo/client/dbclient.h"  // the mongo c++ driver
#include <iostream>
using namespace mongo;
//db-generated object ids of native libs
static OID natlib_oid;
static char appname[MAX_PATH*4]; //utf-16 max. 4 bytes
static char libpath[MAX_PATH*4]; //utf-16 max. 4 bytes

//communicate the object id of the inserted natlib to the insertions of the natlib functions 
static BSONElement be;


static void traverse_binary_functions(boost::shared_ptr<DBClientBase> conn);
static void traverse_binary_entrypoints();
static qflow_chart_t* create_flowchart(func_t *func);
static void delete_flowchart(qflow_chart_t *fc);

FILE *fp;
//SOCKET client_sock;

int __stdcall IDAP_init(void)
{
	/* checks, in cases of failures PLUGIN_SKIP returned*/

	return PLUGIN_KEEP;
}

void __stdcall IDAP_term(void)
{
	// executed when exiting, clean-up jobs here.
	return;
}
// arg -> used in batch analysis to communicate thread-specific worker number to plugin,
// which also defines tempfile names 
void __stdcall IDAP_run(int arg)
{
	char input_file[MAXSTR];

	//error logging file in temporary directory 
	sprintf(errlog_file, "%s\\errlog_%u.tmp", centroid_analysis_tmpdir, arg);

#ifdef CDEBUG
	msg("centroids plugin started ...\n");
#endif

	//#ifdef ELF32ARM_CHECK
	char arch[8];
	char abi[128];
	if ((is_elf() != 1) || (is_elf_32bit() != 1)) {
		//#ifdef CDEBUG
		msg("input_file is not identifiable as elf for arm 32 bit\n");
		//#endif
	}
	if ((get_architecture(arch) == -1) || strcmp(arch, "ARM")) {
		//#ifdef CDEBUG
		msg("input_file is not identifiable as elf for arm 32 bit\n");
		//#endif
	}
	if (elf32arm_getarmabi(abi) < 0) {
		qsnprintf(abi, sizeof(abi) - 1, "unknown");
		//#ifdef CDEBUG
		msg("ubable to identify an ARM ABI for the input file\n");
		//#endif
	}
#ifdef CDEBUG
	msg("input file conforms to 32 bit EFL for ARM, ARM abi according to ELF declarations: %s\n", abi);
#endif
	//#endif


#ifdef RESULTS_TO_FILE
	char results_file[MAXSTR];
	if (qsnprintf(results_file, sizeof(results_file) - 1, "D:\\IDAFiles\\centroidplugin%i_results.txt", arg) <= 0) {
		//if (qsnprintf(results_file, sizeof(results_file) - 1,"X:\\ResultsFiles\\centroidplugin%i_results.txt", arg) <= 0) {
		msg("qsnprintf failed to write char array for results file, exit\n");
		return;
	}
	fp = file_open_append(results_file);
	char stf[MAXSTR];
	if (get_input_file_path(input_file, MAXSTR) > 0) {
		if (qsnprintf(stf, sizeof(stf) - 1, "started the centroids plugin with input file %s\n", input_file) <= 0) {
			msg("qsnprintf failed to write char array\n");
		}
		else {
			file_write_str(fp, stf);
		}
	}
	else{
		if (qsnprintf(stf, sizeof(stf) - 1, "started the centroids plugin, failed to retrieve binary name\n") <= 0) {
			msg("qsnprintf failed to write char array\n");
		}
		else {
			file_write_str(fp, stf);
		}
	}
#endif



#ifdef CDEBUG
	//time measurement
	LARGE_INTEGER frequency;        // ticks per second
	LARGE_INTEGER time_1, time_2;           // ticks
	double elapsedtime_msecs;
	// get ticks per second 
	QueryPerformanceFrequency(&frequency);
	//start timer
	QueryPerformanceCounter(&time_1);
#endif

	//store native lib data to mongo db

	//build the ABI string
	std::string abi_str("ARM");
	abi_str.append(abi);

	//retrieve the appname and the library path from ida caller via files in temp analysis dir
	char commappname_file[MAX_PATH];
	char commlibpath_file[MAX_PATH];
	int num_readbytes;
	//correct comm. files identified by worker number in argument of  ida plugin
	sprintf(commappname_file, "%s\\comm_appname_%u.tmp", centroid_analysis_tmpdir, arg);
	//msg("try to retrieve app name from file %s\n", commappname_file);
	if ((num_readbytes = file_readbytecontents(commappname_file, appname)) == -1) {
		FILE *errlogfile_p = file_open_append(errlog_file);
		if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to retrieve an app's name => library skipped\n") > 0) {
			file_write_str(errlogfile_p, errlog_entry);
		}
		file_close(fp);
		msg("error while trying to store native library data into database,exit\n");
		return;
	}
	appname[num_readbytes] = '\0';
	sprintf(commlibpath_file, "%s\\comm_libpath_%u.tmp", centroid_analysis_tmpdir, arg);
	//msg("try to retrieve lib path from file %s\n", commlibpath_file);
	if ((num_readbytes = file_readbytecontents(commlibpath_file, libpath)) == -1) {
		FILE *errlogfile_p = file_open_append(errlog_file);
		if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to retrieve a native library's path of app %s => library skipped\n", appname) > 0) {
			file_write_str(errlogfile_p, errlog_entry);
		}
		file_close(fp);
		msg("error while trying to store native library data into database,exit\n");
		return;
	}
	libpath[num_readbytes] = '\0';
	//msg("retrieved app name %s and lib path %s from files\n", appname, libpath);

	//sharding : store to the shard determined by the number of functions of the native library
	const char* shard_mongoconnstring;
	int shard_number;
	if (get_func_qty() % 2 == 0) {
		shard_number = 0;
	}
	else {
		shard_number = 1;
	}
	//initialize a mongo client instance 
	mongo::client::GlobalInstance instance;
	if (!instance.initialized()) {
			FILE *errlogfile_p = file_open_append(errlog_file);
			if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to initialize mongo client instance, natlib %s from app %s\n => library skipped\n", libpath, appname) > 0) {
				file_write_str(errlogfile_p, errlog_entry);
			}
			file_close(fp);
			msg("ida plugin failure in initializing a mongo db client driver\n");
			return;
	}

	//find the master of the shard (might change e.g. after reboots of the host) and connect to it
	boost::shared_ptr<DBClientBase> conn;
	bool conn_to_master = false;
	do {
		shard_mongoconnstring = mongo_connstrings[shard_number];
		//get a connection to the a server of the shard and check if it is the master or slave (1x replication)  
		std::string errmsg;
		ConnectionString cs = ConnectionString::parse(shard_mongoconnstring, errmsg);
		if (!cs.isValid()) {
			FILE *errlogfile_p = file_open_append(errlog_file);
			if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to connect to the mongo db shard  %s to find master to store data of natlib %s from app %s in database\n => library skipped\n", shard_mongoconnstring, libpath, appname) > 0) {
				file_write_str(errlogfile_p, errlog_entry);
			}
			file_close(fp);
			msg("failure in establishing a mongo db connection: %s\n", errmsg);
			return;
		}
		boost::shared_ptr<DBClientBase> conn_tmp(cs.connect(errmsg));
		if (!conn_tmp) {
			FILE *errlogfile_p = file_open_append(errlog_file);
			if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to connect to the mongo db shard %s to find master to store data of natlib %s from app %s in database\n => library skipped\n", shard_mongoconnstring, libpath, appname) > 0) {
				file_write_str(errlogfile_p, errlog_entry);
			}
			file_close(fp);
			msg("failure in establishing a mongo db connection: %s\n", errmsg);
			return;
		}
		conn = conn_tmp;
		if (!conn_tmp->isMaster(conn_to_master)) {
			FILE *errlogfile_p = file_open_append(errlog_file);
			if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to connect to the mongo db shard %s to find master to store data of natlib %s from app %s in database\n => library skipped\n", shard_mongoconnstring, libpath, appname) > 0) {
				file_write_str(errlogfile_p, errlog_entry);
			}
			file_close(fp);
			msg("failure in establishing a mongo db connection: %s\n", errmsg);
			return;
		}
		shard_number += 2;
	} while (!conn_to_master);
	shard_number -= 2;

	//create a nat. lib object and insert it, also generate object id via builder to use 
	//it immediately for the manual reference to the natlib in the following natfunc obj creations

	try {

		BSONObj natlib = mongo::BSONObjBuilder().genOID().append("appname",
			appname).append("libpath", libpath).append("numfuncs", get_func_qty()).append("abi", abi)
			.append("cluster_groupid", -1).append("clusterid_cen", -1).append("clusterid_cendash", -1).obj();

		//get the created object id
		natlib.getObjectID(be);
		natlib_oid = be.__oid();

		conn->insert("centroids.natlibs", natlib);

		std::string e = conn->getLastError();
		if (!e.empty()) {
			FILE *errlogfile_p = file_open_append(errlog_file);
			if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to store native library data to the mongo db shard %s, natlib %s from app %s, db exception %s => library skipped\n", shard_mongoconnstring, libpath, appname, e) > 0) {
				file_write_str(errlogfile_p, errlog_entry);
			}
			file_close(fp);
			msg("failure in trying to store native library data to the mongo db\n");
			return;
		}
	}
	catch (DBException& e) {
		FILE *errlogfile_p = file_open_append(errlog_file);
		if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to store native library data to the mongo db shard %s, natlib %s from app %s, db exception %s => library skipped\n", shard_mongoconnstring, libpath, appname, e.toString()) > 0) {
			file_write_str(errlogfile_p, errlog_entry);
		}
		file_close(fp);
		msg("failure in trying to store native library data to the mongo db\n");
		return;
	}

	
	traverse_binary_functions(conn);
	//traverse_binary_entrypoints()

#ifdef CDEBUG
	// stop timer
	QueryPerformanceCounter(&time_2);
	// compute and print the elapsed time in millisec
	elapsedtime_msecs = (time_2.QuadPart - time_1.QuadPart) * 1000.0 / frequency.QuadPart;
	msg("centroids plugin finished, runtime %lf msecs\n", elapsedtime_msecs);
#endif
#ifdef RESULTS_TO_FILE
	if (qsnprintf(stf, sizeof(stf) - 1, "exit the centroids plugin\n") <= 0) {
		msg("qsnprintf failed to write char array\n");
	}
	else {
		file_write_str(fp, stf);
	}
	//file_close(fp);
	eclose(fp);
#endif
	return;
}

// no usage
char IDAP_comment[] = "centroids";
char IDAP_help[] = "centroids";
// name in edit->plugins menu, can be overridden in plugins.cfg 
char IDAP_name[] = "centroids";
char IDAP_hotkey[] = "Alt-2";
// exported object
plugin_t PLUGIN =
{
	IDP_INTERFACE_VERSION, // IDA version plug-in is written for
	0, // flags (see below)
	IDAP_init,
	IDAP_term,
	IDAP_run,
	IDAP_comment, // unused
	IDAP_help,    // unused
	IDAP_name,    // name in edit->plugins menu
	IDAP_hotkey
};

//runs through the functions found by IDA and initiates centroid calculation for each
static void traverse_binary_functions(boost::shared_ptr<DBClientBase> conn) {
	//msg("start searching for the binarys functions ...\n");

	//following traversal of code segments isn't necessary for correct function count 
	/*unsigned int num_code_segs = 0;
	for (int s = 0; s < get_segm_qty(); s++) {
	segment_t *seg = getnseg(s);
	if ((seg != NULL) && (seg->type == SEG_CODE)) {
	//msg("code segment found ...\n");
	num_code_segs++;
	}
	}*/

	//unsigned int num_funcs = 0;
	func_t *func_p;
	qstring func_name;
	bool funcname_retrievable = true;
	ssize_t func_name_length;
	qflow_chart_t *func_cfg;
	cfg_centroid_t func_centroid;
	cfg_centroid_t func_centroid_dash;
	int num_funcs;

	//bulk inserts
	std::vector<BSONObj> bulk_data;
	int actbulksize = 0;
	for (size_t i = 0; i < get_func_qty(); i++) {
		func_p = getn_func(i);
		if (func_p == NULL) {
			FILE *errlogfile_p = file_open_append(errlog_file);
			if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to load a function of natlib %s from app %s => function skipped\n", libpath, appname) > 0) {
				file_write_str(errlogfile_p, errlog_entry);
			}
			file_close(fp);
			msg("failed loading a function, continue with next function\n");
			continue;
		}
		else {
			//msg("function loaded, start address %a, end address %a\n", f->startEA, f->endEA);
		}
		num_funcs++;
		func_name_length = 0;
		func_name_length = get_func_name2(&func_name, func_p->startEA);
		if (func_name_length <= 0) {
			funcname_retrievable = false;
#ifdef RESULTS_TO_FILE
			if (qsnprintf(strf, sizeof(strf) - 1, "function with non-identifiable name ...\n") <= 0) {
				msg("qsnprintf failed to write char array\n");
			}
			else {
				file_write_str(fp, strf);
			}
#endif
			msg("failed querying name of the function\n");
		}
		else {
#ifdef RESULTS_TO_FILE
			if (qsnprintf(strf, sizeof(strf) - 1, "function %s ...\n", func_name.c_str()) <= 0) {
				msg("qsnprintf failed to write char array\n");
			}
			else {
				file_write_str(fp, strf);
			}
#endif
		}
		std::string funcname_str("");
		if (funcname_retrievable) {
			funcname_str.append(func_name.c_str());
		}
		else {
			funcname_str.append("unknown");
		}

#ifdef CDEBUG
		if (!strcmp(func_name.c_str(), "_main")) {
#endif
			func_cfg = create_flowchart(func_p);
			if (func_cfg == NULL) {
#ifdef RESULTS_TO_FILE
				if (qsnprintf(strf, sizeof(strf) - 1, "failed creating the cfg of the function") <= 0) {
					msg("qsnprintf failed to write char array\n");
				}
				else {
					file_write_str(fp, strf);
				}
#endif

				FILE *errlogfile_p = file_open_append(errlog_file);
				if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in creating the cfg of a function named %s in natlib %s from app %s => function skipped\n", funcname_str.c_str(), libpath, appname) > 0) {
					file_write_str(errlogfile_p, errlog_entry);
				}
				file_close(fp);

				msg("failed creating the cfg of the function, continue with next function\n");
				continue;
			}
			if (calc_centroid(func_cfg, &func_centroid, &func_centroid_dash) == -1) {
#ifdef RESULTS_TO_FILE
				if (qsnprintf(strf, sizeof(strf) - 1, "failed calculating the centroid value of the function") <= 0) {
					msg("qsnprintf failed to write char array\n");
				}
				else {
					file_write_str(fp, strf);
				}
#endif

				FILE *errlogfile_p = file_open_append(errlog_file);
				if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in calculating the centroids of a function named %s in natlib %s from app %s => function skipped\n", funcname_str.c_str(), libpath, appname) > 0) {
					file_write_str(errlogfile_p, errlog_entry);
				}
				file_close(fp);

				msg("failed calculating the centroid value of the function, continue with next function\n");
				delete_flowchart(func_cfg);
				continue;
			}
			else {
#ifdef CDEBUG
				msg("calculated centroid value to <%lf,%lf,%lf,%i>\n", func_centroid.cx, func_centroid.cy,
					func_centroid.cz, func_centroid.w_bar);
				msg("calculated centroid_dash value to <%lf,%lf,%lf,%i>\n", func_centroid_dash.cx, func_centroid_dash.cy,
					func_centroid_dash.cz, func_centroid_dash.w_bar);
#endif
#ifdef RESULTS_TO_FILE
				if (qsnprintf(strf, sizeof(strf) - 1, "calculated centroid value to <%lf,%lf,%lf,%i>\n", func_centroid.cx, func_centroid.cy,
					func_centroid.cz, func_centroid.w_bar) <= 0) {
					msg("qsnprintf failed to write char array\n");
				}
				else {
					file_write_str(fp, strf);
				}
				if (qsnprintf(strf, sizeof(strf) - 1, "calculated centroid_dash value to <%lf,%lf,%lf,%i>\n", func_centroid_dash.cx, func_centroid_dash.cy,
					func_centroid_dash.cz, func_centroid_dash.w_bar) <= 0) {
					msg("qsnprintf failed to write char array\n");
				}
				else {
					file_write_str(fp, strf);
				}
#endif
			}

			//store native function data in mongo db only if centroid was successfully calculated

			try {

				BSONObj natfunc = mongo::BSONObjBuilder().append("natlib_id",
					natlib_oid).append("funcname", funcname_str.c_str()).append("cenx", func_centroid.cx).append("ceny", func_centroid.cy).
					append("cenz", func_centroid.cz).append("cenw", func_centroid.w_bar).append("cendashx", func_centroid_dash.cx).append("cendashy", func_centroid_dash.cy).
					append("cendashz", func_centroid_dash.cz).append("cendashw", func_centroid_dash.w_bar).obj();
				//bulk inserts of 1000 funcs 
				bulk_data.push_back(natfunc);
				actbulksize++;

				if (actbulksize % 1000 == 0) {
					conn->insert("centroids.natfuncs", bulk_data);
					std::string e = conn->getLastError();
					if (!e.empty()) {
						FILE *errlogfile_p = file_open_append(errlog_file);
						if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to store nat. library function bulk from natlib %s from app %s in database, db exception %s => bulk skipped\n", libpath, appname, e) > 0) {
							file_write_str(errlogfile_p, errlog_entry);
						}
						file_close(fp);
						msg("failure in trying to store nat. library function data to the mongo db\n");
					}
					bulk_data.clear();
					actbulksize = 0;
				}
			}
			catch (DBException& e) {
				FILE *errlogfile_p = file_open_append(errlog_file);
				if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to store nat. library function bulk from natlib %s from app %s in database, db exception %s => bulk skipped\n", libpath, appname, e.toString()) > 0) {
					file_write_str(errlogfile_p, errlog_entry);
				}
				file_close(fp);
				msg("failure in trying to store nat. library function data to the mongo db\n");
			}

			delete_flowchart(func_cfg);
#ifdef CDEBUG
			break;
		}
#endif
	}
	//insert remaining bulk
	
	if (bulk_data.size() != 0) {
		try {
			conn->insert("centroids.natfuncs", bulk_data);
			//bulk_data.clear();
			//actbulksize = 0;
		 	std::string e = conn->getLastError();
		if (!e.empty()) {
			FILE *errlogfile_p = file_open_append(errlog_file);
			if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to store nat. library function bulk from natlib %s from app %s in database, db exception %s => bulk skipped\n", libpath, appname, e) > 0) {
				file_write_str(errlogfile_p, errlog_entry);
			}
			file_close(fp);
			msg("failure in trying to store nat. library function data to the mongo db\n");
		 }

	    }
	    catch (DBException& e) {
		  FILE *errlogfile_p = file_open_append(errlog_file);
		  if (qsnprintf(errlog_entry, sizeof(errlog_entry) - 1, "ida plugin failure in trying to store nat. library function bulk from natlib %s from app %s in database, db exception %s => bulk skipped\n", libpath, appname, e.toString()) > 0) {
			file_write_str(errlogfile_p, errlog_entry);
		  }
		file_close(fp);
		msg("failure in trying to store nat. library function data to the mongo db\n");
	    }
	} //end if balk_data.size() != 0 
#ifdef RESULTS_TO_FILE
	if (qsnprintf(strf, sizeof(strf) - 1, "%i functions loaded and analyzed\n", num_funcs) <= 0) {
		msg("qsnprintf failed to write char array\n");
	}
	else {
		file_write_str(fp, strf);
	}
#endif
	return;
}

//traverses the  entry points found by IDA
static void traverse_binary_entrypoints() {
	unsigned int num_func_entrypoints = 0;
	unsigned int num_nofunc_entrypoints = 0;
	//msg("start searching for the binarys entrypoints\n");
	size_t num_entrypoints = get_entry_qty();
	for (size_t i = 0; i < num_entrypoints; i++)
	{
		asize_t ord = get_entry_ordinal((int)i);
		ea_t ea = get_entry(ord);
		if (ea == BADADDR) {
			msg("failed loading an entry point, continue searching\n");
			continue;
		}
		else {
			//msg("entry point found at address %a\n", ea);
		}
		char entry_name[MAXSTR];
		if (get_entry_name(ord, entry_name, sizeof(entry_name) == -1)) {
			//msg("failed querying name of the entry point\n");
		}
		else {
			//msg("entry points name is %s\n", entry_name);
		}
		//try to determine if entry point is funtion (or e.g. object)
		if (get_func(ea))
		{
			//func_t *pfn = get_func(ea);
			//msg("the entry point is a function\n");
			num_func_entrypoints++;
			continue;
		}

		//msg("the entry point was not identifiable as a function\n");
		num_nofunc_entrypoints++;
	}
	//msg("finished search for entrypoints - %u entrypoints found, %u identified as functions, %u not\n",
	//num_entrypoints, num_func_entrypoints, num_nofunc_entrypoints);
	return;
}

static qflow_chart_t* create_flowchart(func_t *func) {
	//no external blocks, predecessors needed for loop identification
	//qflow_chart_t *fc = new qflow_chart_t("Flow Chart", NULL, func->startEA, func->endEA, /*FC_NOEXT |*/ FC_PREDS);
	qflow_chart_t *fc = new qflow_chart_t("Flow Chart", func, NULL, NULL,/*FC_APPND |*/ /*FC_NOEXT |*/ FC_PREDS);
	return fc;
}

static void delete_flowchart(qflow_chart_t *fc) {
	delete fc;
}