#include "centroid.h"
#include "cfgtools.h"
#include <vector>
#include <set>

//cfg related data
static qvector<qbasic_block_t>  *cfg_blocks;
static int cfg_num_nodes;
static int *immdoms; //indices and entries IDA node numbers
//not all nodes might be reachable from node 0
static int cfg_num_nodes_reachable;
static bool *reachable_nodes;
//holds for each node a set of its unique loop heads (necessary as loops with same head not nested) 
static std::set<int> *loopheads_by_node;

//centroid related data, numerators and dominators accumulated by node traversal
static int cxnum_acc, cxnum_dash_acc;
static int cynum_acc, cynum_dash_acc;
static int cznum_acc, cznum_dash_acc;
static int wbar_acc, wbar_dash_acc;
static int *seqnums_by_node;
static int *numsuccs_by_node;
static int *numinstructs_by_node;
static int *numcalls_by_node;
static int *label1;
static int *label2;
static int counter;


//accumulates centroid related values per node before final centroid calculation
static int accumulate_for_centroid(qvector<qbasic_block_t>  *blocks, int node, int *visited);

//recursively accumulates  centroid related values resp. a natural loop, loop heads are considered part of their loop
static void acc_cz_in_naturalloop(qvector<qbasic_block_t>  *blocks, int node, int head_node, int* visited);

static int initialize(qflow_chart_t *cfg, cfg_centroid_t* cen, cfg_centroid_t* cen_dash);
static void cleanup();

int calc_centroid(qflow_chart_t *cfg, cfg_centroid_t* cen, cfg_centroid_t* cen_dash) {

	if (initialize(cfg, cen, cen_dash) == -1) return -1;

	//first calculate dominance tree and dfs-reachability of the cfg's nodes
	cfg_dominance(cfg_blocks, immdoms, reachable_nodes, &cfg_num_nodes_reachable);

	//then accumulate centroid related values by dfs
	for (int i = 0; i < cfg_num_nodes; i++) {
		label1[i] = 0;
	}
	counter = 1; //preorder counter, seqnums start at 1 

	if (accumulate_for_centroid(cfg_blocks, 0, label1) == -1) {
		cleanup();
		return -1;
	}

#ifdef CDEBUG
	msg("calculated dfs seqnums, num succs and num instructions:\n");
	for (int i = 0; i < cfg_num_nodes; i++) {
		msg("node %i  :  seqnum %i, numsuccs %i, numinstrs %i\n", i, seqnums_by_node[i],
			numsuccs_by_node[i], numinstructs_by_node[i]);
	}
#endif

	//finally calculate centroid value with the accumulations, denominators might be zero e.g. in case of 
	//cfgs with one node and no edges
	cen->cx = (wbar_acc != 0) ? (((double)cxnum_acc) / ((double)wbar_acc)) : 0.0;
	cen->cy = (wbar_acc != 0) ? (((double)cynum_acc) / ((double)wbar_acc)) : 0.0;
	cen->cz = (wbar_acc != 0) ? (((double)cznum_acc) / ((double)wbar_acc)) : 0.0;
	cen->w_bar = wbar_acc;
	cen_dash->cx = (wbar_dash_acc != 0) ? (((double)cxnum_dash_acc) / ((double)wbar_dash_acc)) : 0.0;
	cen_dash->cy = (wbar_dash_acc != 0) ? (((double)cynum_dash_acc) / ((double)wbar_dash_acc)) : 0.0;
	cen_dash->cz = (wbar_dash_acc != 0) ? (((double)cznum_dash_acc) / ((double)wbar_dash_acc)) : 0.0;
	cen_dash->w_bar = wbar_dash_acc;

	cleanup();
	return 0;
}

static int accumulate_for_centroid(qvector<qbasic_block_t>  *blocks, int node, int *visited) {

	int num_instr_node;
	int num_instr_succnode;
	int num_calls_node;
	int num_calls_succnode;

	visited[node] = 1;
	qbasic_block_t *act_block = &(cfg_blocks->at(node));

#ifdef CDEBUG
	msg("accumulate centroid values: visit node %i\n", node);
#endif

	//set dfs-reachable node's sequence number as dfs number 
	seqnums_by_node[node] = counter;
	counter++;
	//count the instructions of the corresponding basic block
	if (count_bb_instrs(act_block, &(numinstructs_by_node[node]), &(numcalls_by_node[node])) == -1) {
		msg("failed analyzing the instructions of a basic block\n");
		return -1;
	}

	//during dfs-recursion search for natural loops, i.e. if immediate dfs-successor dominates a node
	intseq_t succs = act_block->succ;
	int num_succs = succs.size();
	int succnode;

	//set node's number of successor
	numsuccs_by_node[node] = num_succs;

	for (int i = 0; i < num_succs; i++) {
		succnode = succs[i];
		int dom = node;
		int old_dom;
		//if succ is dom of act node than backward edge found, so traverse immdom tree from act node to root to 
		//check all doms of act node 
		do{
			old_dom = dom;
			if (dom == succnode) {
				//backedge found - accumulate for centroid's cz by traversing all nodes of natural loop 
				for (int j = 0; j < cfg_num_nodes; j++) {
					label2[j] = 0;
				}
				acc_cz_in_naturalloop(cfg_blocks, node, succnode, label2);
				break;
			}
			dom = immdoms[dom];
		} while (dom != old_dom);
		if (visited[succnode] == 0) {
			//dfs recursion
			if (accumulate_for_centroid(blocks, succnode, visited) == -1) return -1;
		}
		//accumulate for centroid's cx and cy values here, successors now processed  
		num_instr_node = numinstructs_by_node[node];
		num_instr_succnode = numinstructs_by_node[succnode];
		num_calls_node = numcalls_by_node[node];
		num_calls_succnode = numcalls_by_node[succnode];
		wbar_acc += (num_instr_node + num_instr_succnode);
		wbar_dash_acc += (num_instr_node + num_calls_node + num_instr_succnode + num_calls_succnode);
		cxnum_acc += (num_instr_node*seqnums_by_node[node] + num_instr_succnode*seqnums_by_node[succnode]);
		cxnum_dash_acc += ((num_instr_node + num_calls_node)*seqnums_by_node[node] + (num_instr_succnode + num_calls_succnode)*seqnums_by_node[succnode]);
		cynum_acc += ((num_instr_succnode*numsuccs_by_node[succnode]) + (num_instr_node*num_succs));
		cynum_dash_acc += (((num_instr_succnode + num_calls_succnode)*numsuccs_by_node[succnode]) + ((num_instr_node + num_calls_node)*num_succs));

	}
	return 0;
}

static void acc_cz_in_naturalloop(qvector<qbasic_block_t>  *blocks, int node, int head_node, int* visited) {

	//reverse DFS search back to loop head collects all loop nodes 
	qbasic_block_t *bb = &(blocks->at(node));
	visited[node] = 1;
	intseq_t npreds;
	int prednode;

	//check if node already part of loop with same head, if so these two loops are not nested => do not count them twice for node
	if (loopheads_by_node[node].count(head_node) == 0) {
		//no => new loop head found for that node
		loopheads_by_node[node].insert(head_node);
		//for all edges the node is part of accumulate for centroids cz numerator
		npreds = bb->pred;
		for (int i = 0; i < npreds.size(); i++) {
			prednode = npreds[i];
			if (reachable_nodes[prednode]) {
				cznum_acc += numinstructs_by_node[node];
				cznum_dash_acc += (numinstructs_by_node[node] + numcalls_by_node[node]);
			}
		}
		cznum_acc += (numinstructs_by_node[node] * bb->succ.size());
		cznum_dash_acc += ((numinstructs_by_node[node] + numcalls_by_node[node]) * (bb->succ.size()));
	}

	//recurse in direction to loop head
	if (node != head_node) {
		intseq_t npreds = bb->pred;
		int prednode;
		for (int i = 0; i < npreds.size(); i++) {
			prednode = npreds[i];
			if ((visited[prednode] == 0) && reachable_nodes[prednode]) { //only go to nodes reachable from root
				acc_cz_in_naturalloop(blocks, prednode, head_node, visited);
			}
		}
	}
}

static int initialize(qflow_chart_t *cfg, cfg_centroid_t* cen, cfg_centroid_t* cen_dash) {
	if ((cfg == NULL) || (cen == NULL) || (cen_dash == NULL)) {
		msg("failed to calculate centroid, bad function parameters\n");
		return -1;
	}
	cfg_num_nodes = cfg->size();
	cfg_blocks = &(cfg->blocks);
	if ((cfg_num_nodes <= 0) || (cfg_blocks->size() < cfg_num_nodes)) {
		msg("failed to retrieve the basic blocks of the function's cfg\n");
		return -1;
	}

	immdoms = new int[cfg_num_nodes];
	reachable_nodes = new bool[cfg_num_nodes];
	seqnums_by_node = new int[cfg_num_nodes];
	numsuccs_by_node = new int[cfg_num_nodes];
	numinstructs_by_node = new int[cfg_num_nodes];
	numcalls_by_node = new int[cfg_num_nodes];
	loopheads_by_node = new std::set<int>[cfg_num_nodes];
	label1 = new int[cfg_num_nodes];
	label2 = new int[cfg_num_nodes];
	cxnum_acc = cynum_acc = cznum_acc = wbar_acc = 0;
	wbar_dash_acc = cxnum_dash_acc = cynum_dash_acc = cznum_dash_acc = 0;
	return 0;
}
static void cleanup() {
	delete[] label1;
	delete[] label2;
	delete[] seqnums_by_node;
	delete[] numsuccs_by_node;
	delete[] numinstructs_by_node;
	delete[] numcalls_by_node;
	delete[] loopheads_by_node;
	delete[] immdoms;
	delete[] reachable_nodes;
}