#include "cfgtools.h"
#include <vector>
#include <set>

static void calc_postorder_and_reachability_dfs(qvector<qbasic_block_t>  *blocks, int node, bool *visited, int *num_reachable_nodes,
	int *nodes_by_postnum, int *postnums_by_node, int * postnum_counter);

void cfg_dominance(qvector<qbasic_block_t>  *cfg_blocks, int *immdoms, bool* reachable_nodes, int *num_nodes_reachable) {
	int cfg_num_nodes = cfg_blocks->size();
	int postnum_counter;
	int local_num_nodes_reachable;
	int *immdoms_by_postnum;
	int *nodes_by_postnum;;
	int *postnums_by_node;
	//po_nums begin at 1, so following two arrays 1 slot larger
	immdoms_by_postnum = new int[cfg_num_nodes + 1];
	nodes_by_postnum = new int[cfg_num_nodes + 1];
	postnums_by_node = new int[cfg_num_nodes];

	//first do a dfs-search to calculate dfs-reachable nodes and their post order numbers 
	(*num_nodes_reachable) = 0;
	for (int i = 0; i < cfg_num_nodes; i++) {
		reachable_nodes[i] = false;
	}
	postnum_counter = 1; //dfs postorder counter, po nums start at 1
	calc_postorder_and_reachability_dfs(cfg_blocks, 0, reachable_nodes, num_nodes_reachable,
		nodes_by_postnum, postnums_by_node, &postnum_counter);
	local_num_nodes_reachable = (*num_nodes_reachable);

#ifdef CDEBUG
	msg("%u nodes reachable, as follows:\n", local_num_nodes_reachable);
	for (int i = 0; i < cfg_num_nodes; i++) {
		msg("node %i: %d\n", i, reachable_nodes[i]);
	}
	msg("... set postorder numbers:\n");
	for (int po_num = 1; po_num <= local_num_nodes_reachable; po_num++) {
		int node = nodes_by_postnum[po_num];
		msg("node %u ^with post order number %i\n", nodes_by_postnum[po_num], po_num);
	}
#endif

	//initialize immdoms array index by postorder numbers, nums beginning with 1 for faster tests for zero 
	//all reachable and unreachable nodes initialize to 0 
	for (int i = 0; i < local_num_nodes_reachable; i++){
		immdoms_by_postnum[i] = 0;
	}
	//root node immdom of itself and has highest post-order number
	immdoms_by_postnum[local_num_nodes_reachable] = local_num_nodes_reachable;

#ifdef CDEBUG
	msg("initial immdoms array for reachable nodes, indexed by post order numbers:\n");
	for (int po_num = 1; po_num <= local_num_nodes_reachable; po_num++) {
		msg("%i ", immdoms_by_postnum[po_num]);
	}
	msg("\n");
#endif

	//something changed in last iteration ?
	bool changed;

	int act_node;
	qbasic_block_t *act_block;
	int new_immdom;
	int pred_ponum;
	int finger1, finger2;

	do {
#ifdef CDEBUG
		msg("start fixpoint iteration\n");
#endif
		changed = false;
		// fetch all nodes in rev. post order, immdom entry of node is found by taking the intersection of its predecessors,
		// process is repeated until no changes were made 
		for (int act_ponum = local_num_nodes_reachable - 1; act_ponum >= 1; act_ponum--) {
			act_node = nodes_by_postnum[act_ponum];
			act_block = &(cfg_blocks->at(act_node));

			// first set immdom of actnode  = first processed reachable predecessor of actnode 
			//(at least one processed cause of its higher postorder number) 
			intseq_t preds = act_block->pred;
			int i = 0;
			int act_pred;
			for (; i < preds.size(); i++) {
				act_pred = preds[i];
				if (!reachable_nodes[act_pred]) continue;
				pred_ponum = postnums_by_node[act_pred];
				if (immdoms_by_postnum[pred_ponum] != 0) { //processed (in any iteration) pred ?
					new_immdom = pred_ponum;
					break;
				}
			}
			// then for all other already processed predecessors do an intersection
			for (; i < preds.size(); i++) {
				act_pred = preds[i];
				if (!reachable_nodes[act_pred]) continue;
				pred_ponum = postnums_by_node[act_pred];
				if (immdoms_by_postnum[pred_ponum] != 0) {//processed (in any iteration) pred ?
					// intersection: 
					// traverse currently set immdom of actnode and immdom of predecessor backwards to root until 
					// common immdom found, that's the resulting immdom of actnode
					finger1 = pred_ponum;
					finger2 = new_immdom;
					while (finger1 != finger2) {
						while (finger1 < finger2) finger1 = immdoms_by_postnum[finger1];
						while (finger2 < finger1) finger2 = immdoms_by_postnum[finger2];
					}
					new_immdom = finger1;
				}
			}
			//finally for that node check if the new immdom represents a change 
			if (immdoms_by_postnum[act_ponum] != new_immdom) {
				immdoms_by_postnum[act_ponum] = new_immdom;
				changed = true;
			}

		} //end for all nodes in one iteration
#ifdef CDEBUG
		msg("... fixpoint iteration finished, new immdoms array for reachable nodes ind. by po numbers:\n");
		for (int po_num = 1; po_num <= local_num_nodes_reachable; po_num++) {
			msg("%i ", immdoms_by_postnum[po_num]);
		}
		msg("\n");
#endif
	} while (changed);
	//end dom algo, now fill immdoms by node out param, unreachable nodes immdoms of themselfs
	for (int node = 0; node < cfg_num_nodes; node++) {
		if (reachable_nodes[node]) {
			immdoms[node] = nodes_by_postnum[immdoms_by_postnum[postnums_by_node[node]]];
		}
		else {
			immdoms[node] = node;
		}
	}
#ifdef CDEBUG
	msg("... calculated immdoms tree array ind. by nodes:\n");
	for (int node = 0; node < cfg_num_nodes; node++) {
		msg("%i ", immdoms[node]);
	}
	msg("\n");
#endif

	delete[] immdoms_by_postnum;
	delete[] nodes_by_postnum;
	delete[] postnums_by_node;

}

//dfs post order number and reachability calculation
static void calc_postorder_and_reachability_dfs(qvector<qbasic_block_t>  *blocks, int node, bool *visited, int *num_reachable_nodes,
	int *nodes_by_postnum, int *postnums_by_node, int * postnum_counter){

	//mark as visited/update reachability info
	visited[node] = true;
	(*num_reachable_nodes) = (*num_reachable_nodes) + 1;

	//recurse
	qbasic_block_t *act_block = &(blocks->at(node));
	intseq_t succs = act_block->succ;
	int num_succs = succs.size();
	int succnode;
	for (int i = 0; i < num_succs; i++) {
		succnode = succs[i];
		if (visited[succnode] == 0) {
			calc_postorder_and_reachability_dfs(blocks, succnode, visited, num_reachable_nodes,
				nodes_by_postnum, postnums_by_node, postnum_counter);
		}
	}

	//set postorder number of node here, as all succs now processed
	nodes_by_postnum[(*postnum_counter)] = node;
	postnums_by_node[node] = (*postnum_counter);
	(*postnum_counter) = (*postnum_counter) + 1;
}

//counts the number of instructions and calls in a cfg basic block
int count_bb_instrs(qbasic_block_t* bb, int *num_instr, int *num_calls) {

#ifdef CDEBUG
	msg("start counting instructions of basic block, address range %a - %a ...\n", bb->startEA, bb->endEA);
#endif

	int num_bb_instrs = 0;
	int num_bb_calls = 0;


	/*no empty bbs should exist,
	traverse bb via start addresses of "items" (code or data, see bytes.hpp),
	endEA should be adress !after bb w.r.t.area.h ("excluded", however my tests found the following:
	. both for x86 and arm functions with startEA = endEA, so use "<= endEA" for loop test to avoid counting 0 instructions fpr bbs
	- for Android- ELF for AARM blocks with 1 instruction (instr. BX) with endEA = startEA + 4, so with "<=" wrongly instruction count 2
	=> TODO: how to code loop test (one possiblity use test_ea_end of SDK-API?, if wrong however: what's better for centroid calculation) ?*/
	for (ea_t addr = bb->startEA; addr <= bb->endEA;) {
		flags_t flags = getFlags(addr);
		if (flags == 0) {
			msg("failed to retrieve the flags of a basic block item's start address\n");
			return -1;
		}
		// check if we are at head byte of code, i.e. start byte of instruction 
		if (isHead(flags) && isCode(flags)) {
			num_bb_instrs++;
			//is instruction a call ?
			char mnem[MAXSTR];
			// fill the cmd structure with the disassembly of the current address and get the mnemonic text
			if (ua_mnem(addr, mnem, sizeof(mnem) - 1) == NULL) {
				msg("failed to decode the mnemonics of a basic block's instruction");
				return -1;
			}
			// check the mnemonic of the address against call mnemonics
			if (cmd.itype == NN_call || cmd.itype == NN_callfi || cmd.itype == NN_callni)
				num_bb_calls++;
		}
		addr = next_head(addr, 0xffffffff);
		//check w.r.t. IDAs address space boundary
		if (addr == BADADDR) break;
	}
	//set out parameters
	(*num_instr) = num_bb_instrs;
	(*num_calls) = num_bb_calls;
	return 0;
}
