#include "binaryinfo.h"

//elf file carving with elfio
#include <elfio/elfio.hpp>
#include <elfio/elf_types.hpp>
#include <elfio/elfio_symbols.hpp>
#include <elfio/elfio_section.hpp>

//used to carve ARM attributes section of ELF files
static unsigned char *carve_for_armabi(unsigned char *p, bool *armabi_found, char abi[128]);
//carves an uleb128 encoded value
static int read_leb128(unsigned char *p, unsigned int *bytelength);

#define ARRAY_SIZE(array) (sizeof(array) / sizeof(*array))
typedef struct
{
	int tag;
	const char *name;
	/* 0 = special, 1 = string, 2 = uleb123, > 0x80 == table lookup.  */
	int type;
	const char **table;
} arm_attr_public_tag;
static const char *arm_attr_tag_CPU_arch[] =
{ "Pre-v4", "v4", "v4T", "v5T", "v5TE", "v5TEJ", "v6", "v6KZ", "v6T2", "v6K", "v7" };
static const char *arm_attr_tag_ARM_ISA_use[] = { "No", "Some" };
static const char *arm_attr_tag_THUMB_ISA_use[] = { "No", "Thumb-1", "Thumb-2" };
static const char *arm_attr_tag_VFP_arch[] = { "No", "VFPv1", "VFPv2" };
static const char *arm_attr_tag_WMMX_arch[] = { "No", "WMMXv1" };
static const char *arm_attr_tag_NEON_arch[] = { "No", "NEONv1" };
static const char *arm_attr_tag_ABI_PCS_config[] = { "None", "Bare platform", "Linux application", "Linux DSO", "PalmOS 2004",
"PalmOS (reserved)", "SymbianOS 2004", "SymbianOS (reserved)" };
static const char *arm_attr_tag_ABI_PCS_R9_use[] = { "V6", "SB", "TLS", "Unused" };
static const char *arm_attr_tag_ABI_PCS_RW_data[] = { "Absolute", "PC-relative", "SB-relative", "None" };
static const char *arm_attr_tag_ABI_PCS_RO_DATA[] = { "Absolute", "PC-relative", "None" };
static const char *arm_attr_tag_ABI_PCS_GOT_use[] = { "None", "direct", "GOT-indirect" };
static const char *arm_attr_tag_ABI_PCS_wchar_t[] = { "None", "??? 1", "2", "??? 3", "4" };
static const char *arm_attr_tag_ABI_FP_rounding[] = { "Unused", "Needed" };
static const char *arm_attr_tag_ABI_FP_denormal[] = { "Unused", "Needed" };
static const char *arm_attr_tag_ABI_FP_exceptions[] = { "Unused", "Needed" };
static const char *arm_attr_tag_ABI_FP_user_exceptions[] = { "Unused", "Needed" };
static const char *arm_attr_tag_ABI_FP_number_model[] = { "Unused", "Finite", "RTABI", "IEEE 754" };
static const char *arm_attr_tag_ABI_align8_needed[] = { "No", "Yes", "4-byte" };
static const char *arm_attr_tag_ABI_align8_preserved[] = { "No", "Yes, except leaf SP", "Yes" };
static const char *arm_attr_tag_ABI_enum_size[] = { "Unused", "small", "int", "forced to int" };
static const char *arm_attr_tag_ABI_HardFP_use[] = { "As Tag_VFP_arch", "SP only", "DP only", "SP and DP" };
static const char *arm_attr_tag_ABI_VFP_args[] = { "AAPCS", "VFP registers", "custom" };
static const char *arm_attr_tag_ABI_WMMX_args[] = { "AAPCS", "WMMX registers", "custom" };
static const char *arm_attr_tag_ABI_optimization_goals[] = { "None", "Prefer Speed", "Aggressive Speed", "Prefer Size",
"Aggressive Size", "Prefer Debug", "Aggressive Debug" };

static const char *arm_attr_tag_ABI_FP_optimization_goals[] = { "None", "Prefer Speed", "Aggressive Speed", "Prefer Size",
"Aggressive Size", "Prefer Accuracy", "Aggressive Accuracy" };
#define LOOKUP(id, name) \
  {id, #name, 0x80 | ARRAY_SIZE(arm_attr_tag_##name), arm_attr_tag_##name}
static arm_attr_public_tag arm_attr_public_tags[] =
{
	{ 4, "CPU_raw_name", 1, NULL },
	{ 5, "CPU_name", 1, NULL },
	LOOKUP(6, CPU_arch),
	{ 7, "CPU_arch_profile", 0, NULL },
	LOOKUP(8, ARM_ISA_use),
	LOOKUP(9, THUMB_ISA_use),
	LOOKUP(10, VFP_arch),
	LOOKUP(11, WMMX_arch),
	LOOKUP(12, NEON_arch),
	LOOKUP(13, ABI_PCS_config),
	LOOKUP(14, ABI_PCS_R9_use),
	LOOKUP(15, ABI_PCS_RW_data),
	LOOKUP(16, ABI_PCS_RO_DATA),
	LOOKUP(17, ABI_PCS_GOT_use),
	LOOKUP(18, ABI_PCS_wchar_t),
	LOOKUP(19, ABI_FP_rounding),
	LOOKUP(20, ABI_FP_denormal),
	LOOKUP(21, ABI_FP_exceptions),
	LOOKUP(22, ABI_FP_user_exceptions),
	LOOKUP(23, ABI_FP_number_model),
	LOOKUP(24, ABI_align8_needed),
	LOOKUP(25, ABI_align8_preserved),
	LOOKUP(26, ABI_enum_size),
	LOOKUP(27, ABI_HardFP_use),
	LOOKUP(28, ABI_VFP_args),
	LOOKUP(29, ABI_WMMX_args),
	LOOKUP(30, ABI_optimization_goals),
	LOOKUP(31, ABI_FP_optimization_goals),
	{ 32, "compatibility", 0, NULL } };
#undef LOOKUP

int get_architecture(char arch[8]) {
	char input_file[MAXSTR];
	if (get_input_file_path(input_file, MAXSTR) <= 0) {
		msg("failed to retrieve input file name\n");
		return -1;
	}
	inf.get_proc_name(arch);
	return 0;
}

//checks if IDA input file is ELF format
int is_elf() {
	using namespace ELFIO;
	char input_file[MAXSTR];
	if (get_input_file_path(input_file, MAXSTR) <= 0) {
		msg("failed to retrieve input file name\n");
		return -1;
	}
	elfio reader;
	if (!reader.load(input_file)) {
		return 0;
	}
	else {
		return 1;
	}
}

int is_elf_32bit() {
	using namespace ELFIO;
	char input_file[MAXSTR];
	if (get_input_file_path(input_file, MAXSTR) <= 0) {
		msg("failed to retrieve input file name\n");
		return -1;
	}
	elfio reader;
	if (!reader.load(input_file)) {
		return 0;
	}
	else {
		if (reader.get_class() == ELFCLASS32) {
			return 1;
		}
		return 0;
	}
}

int is_elf_64bit() {
	using namespace ELFIO;
	char input_file[MAXSTR];
	if (get_input_file_path(input_file, MAXSTR) <= 0) {
		msg("failed to retrieve input file name\n");
		return -1;
	}
	elfio reader;
	if (!reader.load(input_file)) {
		return 0;
	}
	else {
		if (reader.get_class() == ELFCLASS64) {
			return 1;
		}
		return 0;
	}
}

int elf32arm_getarmabi(char abi[128]) {

	int out_p = 0;
	unsigned char *p;
	unsigned char *end;
	unsigned int section_len;
	unsigned int len;
	unsigned i;
	using namespace ELFIO;
	uint const* temp;
	char input_file[MAXSTR];
	char stf[MAXSTR];
	if (get_input_file_path(input_file, MAXSTR) <= 0) {
		msg("failed to retrieve input file name\n");
		return -1;
	}

	//check for ELF for ARM 32 bit input file format

	inf.get_proc_name(stf);
	stf[8] = 0;
	if (strcmp("ARM", stf)) {
		msg("unable to process input file as ELF for ARM 32 bit\n");
		return -1;
	}
	elfio reader;
	if (!reader.load(input_file)) {
		msg("unable to process input file as ELF for ARM 32 bit\n");
		return -1;
	}
	if (reader.get_class() != ELFCLASS32) {
		msg("unable to process input file as ELF for ARM 32 bit\n");
		return -1;
	}

	bool elf_encoding_littleendian = true;
	if (reader.get_encoding() != ELFDATA2LSB)
		elf_encoding_littleendian = false;

	//try to localize .ARM.attributes section 
	Elf_Half sec_num = reader.sections.size();
	for (int i = 0; i < sec_num; ++i) {
		if (!strcmp(".ARM.attributes", reader.sections[i]->get_name().c_str())) {
			//msg("ARM attributes section found\n");
			unsigned int sec_size = reader.sections[i]->get_size();
			unsigned char *p = (unsigned char *)reader.sections[i]->get_data();
			if (!p) continue;
			if (*p == 'A') //skip 'A' byte
			{
				len = sec_size - 1;
				p++;
				while (len > 0)
				{
					//carve a section
					int namelen;
					bool public_section;
					//carve section length
					section_len = *((unsigned int *)p);
					if (!elf_encoding_littleendian) {
						temp = reinterpret_cast<uint const*>(p);
						section_len = (*temp >> 24) | ((*temp & 0x00ff0000) >> 8) | ((*temp & 0x0000ff00) << 8) | (*temp << 24);
					}
					p += 4;
					if (section_len > len)
					{
						msg("bad ELF section length (%d > %d)\n", (int)section_len, (int)len);
						return -1;
						//section_len = len;
					}
					len -= section_len;
					//carve section name
					//msg("Section: %s\n", p);
					if (strcmp((char *)p, "aeabi") == 0)
						public_section = true;
					else
						public_section = false;
					namelen = strlen((char *)p) + 1;
					p += namelen;
					//carve subsections
					section_len -= namelen + 4;
					while (section_len > 0)
					{
						//carve subsection length
						int tag = *(p++);
						int val;
						unsigned int size;
						size = *((unsigned int *)p);
						if (!elf_encoding_littleendian) {
							temp = reinterpret_cast<uint const*>(p);
							size = (*temp >> 24) | ((*temp & 0x00ff0000) >> 8) | ((*temp & 0x0000ff00) << 8) | (*temp << 24);
						}
						if (size > section_len)
						{
							msg("bad subsection length (%d > %d)\n", (int)size, (int)section_len);
							return -1;
							//size = section_len;
						}
						section_len -= size;
						end = p + size - 1;
						p += 4;
						//carve attributes type
						switch (tag)
						{
						case 1:
							//msg("File Attributes\n");
							break;
						case 2:
							//msg("Section Attributes:");
							goto do_numlist;
						case 3:
							//msg("Symbol Attributes:");
						do_numlist :
							for (;;)
							{
								unsigned int i;
								val = read_leb128(p, &i);
								p += i;
								if (val == 0) break;
								//msg(" %d", val);
							}
								   //msg("\n");
								   break;
						default:
							//in case of unknown tags do nothing
							//msg("Unknown tag: %d\n", tag);
							public_section = false;
							break;
						}
						if (public_section) //in case of "aeabi" section, carve ARM attributs and search for TAG_CPU_arch
						{
							bool armabi_found = false;
							while (p < end) {
								p = carve_for_armabi(p, &armabi_found, abi);
								if (armabi_found) return 0; //if TAG_CPU_arch was found, return immediately
							}
						}
						else
						{
							p = end;
						}
					} //end while section_len > 0
				} // end while len > 0
			} //end if (*p == 'A')
			else
			{
				; //do nothing
			}
		} // end if (!strcmp(".ARM.attributes", reader.sections[i]->get_name().c_str()))
	} //end for (int i = 0; i < sec_num; ++i)
	//no errors, but arm abi wasn't found
	qstrncpy(abi, "unknown", 128);
	return 0;
}


static unsigned char *carve_for_armabi(unsigned char *p, bool *armabi_found, char abi[128])
{
	int tag;
	unsigned int len;
	int val;
	arm_attr_public_tag *attr;
	unsigned i;
	int type;
	tag = read_leb128(p, &len);
	p += len;
	attr = NULL;
	for (i = 0; i < ARRAY_SIZE(arm_attr_public_tags); i++)
	{
		if (arm_attr_public_tags[i].tag == tag)
		{
			attr = &arm_attr_public_tags[i];
			break;
		}
	}
	if (attr)
	{
		//msg("  Tag_%s: ", attr->name);
		switch (attr->type)
		{
		case 0:
			//attribute value type: special
			switch (tag)
			{
			case 7: /* Tag_CPU_arch_profile.  */
				val = read_leb128(p, &len);
				p += len;
				/*switch (val)
				{
				case 0: msg("None\n"); break;
				case 'A': msg("Application\n"); break;
				case 'R': msg("Realtime\n"); break;
				case 'M': msg("Microcontroller\n"); break;
				default: msg("??? (%d)\n", val); break;
				}*/
				break;
			case 30: /* Tag_compatibility.  */
				val = read_leb128(p, &len);
				p += len;
				//msg("flag = %d, vendor = %s\n", val, p);
				p += strlen((char *)p) + 1;
				break;
			default:
				abort();
			}
			return p;
		case 1:
		case 2:
			type = attr->type;
			break;
		default:
			//attribute value type: table lookup
			assert(attr->type & 0x80);
			val = read_leb128(p, &len);
			p += len;
			type = attr->type & 0x7f;
			if (val >= type)
				//msg("??? (%d)\n", val);
				; //do nothing
			else {
				//msg("%s\n", attr->table[val]);
				if (tag == 6) { //CPU_arch tag found ?
					qstrncpy(abi, attr->table[val], 128);
					*armabi_found = true;
				}
			}
			return p;
		}
	}
	else
	{
		if (tag & 1)
			type = 1; /* String.  */
		else
			type = 2; /* uleb128.  */
		//msg("  Tag_unknown_%d: ", tag);
	}
	if (type == 1)
	{
		//attribute value type: string
		//msg("\"%s\"\n", p);
		p += strlen((char *)p) + 1;
	}
	else
	{
		//attribute value type: uleb128
		val = read_leb128(p, &len);
		p += len;
		//msg("%d (0x%x)\n", val, val);
	}
	return p;
}

static int read_leb128(unsigned char * p, unsigned int *bytelength) {
	uint64_t bit = 0;
	int result = 0;
	*bytelength = 0;
	unsigned char * p_old;
	do  {
		*bytelength = *bytelength + 1;
		p_old = p;
		uint64_t b;
		b = *p & 0x7f;

		if (bit >= 64 || b << bit >> bit != b)
			result = (uint64_t)-1;
		else
			result |= b << bit, bit += 7;
		p = p++;

	} while (*(p_old) >= 0x80);
	return result;
}