#include "centroid.h"
#include "cfgtools.h"
#include <vector>
#include <set>

//cfg related data
static qvector<qbasic_block_t>  *cfg_blocks;
static int cfg_num_nodes;
static int *immdoms; //indices and entries IDA node numbers
static int *postdoms; //indices and entries IDA node numbers

//not all nodes might be dfs-reachable from node 0 or reverse dfs-reachable from exit node 
static int num_nodes_dfs_reachable_from_root;
static int num_nodes_revdfs_reachable_from_exit;
static bool *dfs_reachable_from_root;
static bool *revdfs_reachable_from_exit;

//holds for each node a set of its unique loop heads (necessary as loops with same head not nested) 
static std::set<int> *loopheads_by_node;

//centroid related data, numerators and dominators accumulated by node traversal
static int cxnum_acc, cxnum_dash_acc;
static int cynum_acc, cynum_dash_acc;
static int cznum_acc, cznum_dash_acc;
static int wbar_acc, wbar_dash_acc;
static int *seqnums_by_node;
static int *numsuccs_by_node;
static int *numinstructs_by_node;
static int *numcalls_by_node;
static int *loopdepths_by_node;
static int *label1;
static int *label2;
static int counter;

//accumulates centroid related values by cfg traversal, variant with dfs seqnums for
//cfgs without symmetric reachability
static int accumulate_for_centroid_dfsseqnums(qvector<qbasic_block_t>  *blocks, int node, int *visited);
//variant with postdominance analysis based seqnums for common cfgs
static int accumulate_for_centroid_pdomseqnums(qvector<qbasic_block_t>  *blocks, int *postdoms, int node,
	int act_endofbranch, int *visited);
//collects the information by post dominance based traversal necessary to accumulate for centroid related values 
static int accumulate_for_centroid_pdomseqnums_collect(qvector<qbasic_block_t>  *blocks, int *postdoms, int node,
	int act_endofbranch, int *visited);
//recursively traverses a natural loop, loop heads are considered part of their loop,
//if accumulate param true accumulates for cz, else only collects loop depth information
static void cen_analyze_naturalloop(qvector<qbasic_block_t>  *blocks, int node, int head_node, int* visited, bool accumulate);

static int initialize(qflow_chart_t *cfg, cfg_centroid_t* cen, cfg_centroid_t* cen_dash);
static void cleanup();

int calc_centroid(qflow_chart_t *cfg, cfg_centroid_t* cen, cfg_centroid_t* cen_dash) {

	if (initialize(cfg, cen, cen_dash) == -1) return -1;

	//first calculate dominance tree and dfs-reachability of the cfg's nodes
	analyze_dominance(cfg_blocks, immdoms, dfs_reachable_from_root, &num_nodes_dfs_reachable_from_root);


	//then add an exit node, calculate postdominance tree and check if nodes dfs-reachable from 
	//node 0 are the same as the nodes reachable from exit node (pathological case: node with edge
	//to itself => no exit => exit node not logically appendable), if symmetric reachability of cfg is 
	//given calculate sequence numbers with postdominance, else use dfs 

	if (append_exitnode(cfg, dfs_reachable_from_root) == -1) {
		msg("failed to add exit node\n");
		cleanup();
		return -1;
	}

	bool cfg_symmetric_reachability =
		(analyze_postdominance_from_exitnode(cfg_blocks, dfs_reachable_from_root, postdoms, revdfs_reachable_from_exit, 
		&num_nodes_revdfs_reachable_from_exit) == -1) ? false : true;
	
	for (int i = 0; i <= cfg_num_nodes; i++) {
		label1[i] = 0;
	}
	counter = 1; 
	if (cfg_symmetric_reachability) {
		#ifdef CDEBUG
			msg("accumulate for centroid values with postdominance based seqnums\n");
		#endif
		if (accumulate_for_centroid_pdomseqnums(cfg_blocks, postdoms, 0, cfg_num_nodes, label1) == -1) {
			cleanup();
			return -1;
		}
	}
	else {
		#ifdef CDEBUG
			msg("accumulate for centroid values with dfs seqnums\n");
		#endif
		if (accumulate_for_centroid_dfsseqnums(cfg_blocks, 0, label1) == -1) {
			cleanup();
			return -1;
		}
	}

	#ifdef CDEBUG
		msg("calculated dfs seqnums, num succs and num instructions as follows:\n");
		for (int i = 0; i < cfg_num_nodes; i++) {
			msg("node %i  :  seqnum %i, numsuccs %i, numinstrs %i\n", i, seqnums_by_node[i],
			numsuccs_by_node[i], numinstructs_by_node[i]);
		}
	#endif

	//finally calculate centroid value with the accumulations, denominators might be zero e.g. in case of 
	//cfgs with one node and no edges
	cen->cx = (wbar_acc != 0) ? (((double)cxnum_acc) / ((double)wbar_acc)) : 0.0;
	cen->cy = (wbar_acc != 0) ? (((double)cynum_acc) / ((double)wbar_acc)) : 0.0;
	cen->cz = (wbar_acc != 0) ? (((double)cznum_acc) / ((double)wbar_acc)) : 0.0;
	cen->w_bar = wbar_acc;
	cen_dash->cx = (wbar_dash_acc != 0) ? (((double)cxnum_dash_acc) / ((double)wbar_dash_acc)) : 0.0;
	cen_dash->cy = (wbar_dash_acc != 0) ? (((double)cynum_dash_acc) / ((double)wbar_dash_acc)) : 0.0;
	cen_dash->cz = (wbar_dash_acc != 0) ? (((double)cznum_dash_acc) / ((double)wbar_dash_acc)) : 0.0;
	cen_dash->w_bar = wbar_dash_acc;

	cleanup();
	return 0;
}

static int accumulate_for_centroid_dfsseqnums(qvector<qbasic_block_t>  *blocks, int node, int *visited) {

	int num_instr_node;
	int num_instr_succnode;
	int num_calls_node;
	int num_calls_succnode;

	visited[node] = 1;
	qbasic_block_t *act_block = &(cfg_blocks->at(node));

	//set dfs-reachable node's sequence number as dfs number 
	seqnums_by_node[node] = counter;
	counter++;
	//count the instructions of the corresponding basic block
	if (count_basicblock_instructions(act_block, &(numinstructs_by_node[node]), &(numcalls_by_node[node])) == -1) {
		msg("failed analyzing the instructions of a basic block\n");
		return -1;
	}

	//during dfs-recursion search for natural loops, i.e. if immediate dfs-successor dominates a node
	intseq_t succs = act_block->succ;
	int num_succs = succs.size();
	int succnode;

	//calculate node's total number of successors, don't count exit node as succ
	for (int i = 0; i < num_succs; i++) {
		if (succs[i] != (blocks->size() - 1)) numsuccs_by_node[node]++;
	}

	for (int i = 0; i < num_succs; i++) {
		if (succnode == blocks->size() - 1) continue; //no need to analyze edge to exit node
		succnode = succs[i];
		int dom = node;
		int old_dom;
		//if succ is dom of act node than backward edge found, so traverse immdom tree from act node to root to 
		//check all doms of act node 
		do{
			old_dom = dom;
			if (dom == succnode) {
				//backedge found - accumulate for centroid's cz by traversing all nodes of natural loop 
				for (int j = 0; j < cfg_num_nodes; j++) {
					label2[j] = 0;
				}
				cen_analyze_naturalloop(cfg_blocks, node, succnode, label2, true);
				break;
			}
			dom = immdoms[dom];
		} while (dom != old_dom);
		if (visited[succnode] == 0) {
			//dfs recursion
			if(accumulate_for_centroid_dfsseqnums(blocks, succnode, visited) == -1) return -1;
		}
		//accumulate for centroid's cx and cy values here, successors now processed 
		num_instr_node = numinstructs_by_node[node];
		num_instr_succnode = numinstructs_by_node[succnode];
		num_calls_node = numcalls_by_node[node];
		num_calls_succnode = numcalls_by_node[succnode];
		wbar_acc += (num_instr_node + num_instr_succnode);
		wbar_dash_acc += (num_instr_node + num_calls_node + num_instr_succnode + num_calls_succnode);
		cxnum_acc += (num_instr_node*seqnums_by_node[node] + num_instr_succnode*seqnums_by_node[succnode]);
		cxnum_dash_acc += ((num_instr_node + num_calls_node)*seqnums_by_node[node] + (num_instr_succnode + num_calls_succnode)*seqnums_by_node[succnode]);
		cynum_acc += ((num_instr_succnode*numsuccs_by_node[succnode]) + (num_instr_node*num_succs));
		cynum_dash_acc += (((num_instr_succnode + num_calls_succnode)*numsuccs_by_node[succnode]) + ((num_instr_node + num_calls_node)*num_succs));

	}
	return 0;
}

static int accumulate_for_centroid_pdomseqnums(qvector<qbasic_block_t>  *blocks, int *postdoms, int node,
	int act_endofbranch, int *visited) {

	//first collect the values necessary for accumulation via posdominance analysis based cfg traversal
	if (accumulate_for_centroid_pdomseqnums_collect(blocks, postdoms, node, act_endofbranch, visited) == -1) return -1;
	//then accumulate
	qbasic_block_t *act_block;
	intseq_t succs;
	int num_succs, succnode, num_instr_node, num_instr_succnode, num_calls_node, num_calls_succnode;
	for (int node = 0; node < blocks->size() - 1; node++) {
		if (!dfs_reachable_from_root[node]) continue; //don't accumulate for nodes not dfs-reachable 
		act_block = &(cfg_blocks->at(node));
		succs = act_block->succ;
		num_succs = succs.size();
		for (int s = 0; s < num_succs; s++) {
			succnode = succs[s];
			if (succnode == blocks->size() - 1) continue; //don't count edges to exit node
			num_instr_node = numinstructs_by_node[node];
			num_instr_succnode = numinstructs_by_node[succnode];
			num_calls_node = numcalls_by_node[node];
			num_calls_succnode = numcalls_by_node[succnode];
			wbar_acc += (num_instr_node + num_instr_succnode);
			wbar_dash_acc += (num_instr_node + num_calls_node + num_instr_succnode + num_calls_succnode);
			cxnum_acc += (num_instr_node*seqnums_by_node[node] + num_instr_succnode*seqnums_by_node[succnode]);
			cxnum_dash_acc += ((num_instr_node + num_calls_node)*seqnums_by_node[node] + (num_instr_succnode + num_calls_succnode)*seqnums_by_node[succnode]);
			cynum_acc += ((num_instr_succnode*numsuccs_by_node[succnode]) + (num_instr_node*numsuccs_by_node[node]));
			cynum_dash_acc += (((num_instr_succnode + num_calls_succnode)*numsuccs_by_node[succnode]) + ((num_instr_node + num_calls_node)*numsuccs_by_node[node]));
			cznum_acc += ((num_instr_succnode*loopdepths_by_node[succnode]) + (num_instr_node*loopdepths_by_node[node]));
			cznum_dash_acc += (((num_instr_succnode + num_calls_succnode)*loopdepths_by_node[succnode]) + ((num_instr_node + num_calls_node)*loopdepths_by_node[node]));
		}
	}
		return 0;
}

static int accumulate_for_centroid_pdomseqnums_collect(qvector<qbasic_block_t>  *blocks, int *postdoms, int node, 
	int act_endofbranch, int *visited) {

	#ifdef CDEBUG
		msg("collect info for centroid via postdominance, visit node %i, act endofbranch at %i\n", node, act_endofbranch);
	#endif
	//if end of branch to investigate (defined by postdom) reached: do nothing and return 
	//(exit node will thus never be processed)
	if (node == act_endofbranch) {
		#ifdef CDEBUG
			msg("node is act endofbranch, so do not yet visit and return\n");
		#endif
		return 0;
	}

	//mark as visited and give it next sequence number
	visited[node] = 1;
	seqnums_by_node[node] = counter;
	#ifdef CDEBUG
		msg("node %i's seqnum set to %i\n", node, counter);
	#endif
	counter++;

	//count the instructions of the corresponding basic block
	qbasic_block_t *act_block = &(blocks->at(node));
	//number instructions of that node might already have been calculated during branch ordering recursion
	if (numinstructs_by_node[node] == -1) {
		if (count_basicblock_instructions(act_block, &(numinstructs_by_node[node]), &(numcalls_by_node[node])) == -1) {
			msg("failed analyzing the instructions of a basic block\n");
			return -1;
		}
	}

	//recurse
	intseq_t succs = act_block->succ;
	int num_succs = succs.size();
	int succnode;
	qvector<int> unvisited_succs;

	//calculate node's total number of successors, don't count exit node as succ
	for (int i = 0; i < num_succs; i++) {
		if (succs[i] != (blocks->size() - 1)) numsuccs_by_node[node]++;
	}

	for (int i = 0; i < num_succs; i++) {
		succnode = succs[i];
		//identify natural loops via back edges 
		int dom = node;
		int old_dom;
		//if succ is dom of act node than backward edge found, so traverse immdom tree from act node to root to 
		//check all doms of act node 
		do{
			old_dom = dom;
			if (dom == succnode) {
				//backedge found - update loop depths of all nodes in natural loop 
				for (int j = 0; j < cfg_num_nodes; j++) {
					label2[j] = 0;
				}
				cen_analyze_naturalloop(cfg_blocks, node, succnode, label2, false);
				break;
			}
			dom = immdoms[dom];
		} while (dom != old_dom);

		//collect unvisited succs for postdominance based recursion
		if (visited[succnode] == 0)
			unvisited_succs.push_back(succnode);
	}
	int num_unv_succs = unvisited_succs.size();
	#ifdef CDEBUG
		msg("%i currently unvisited successors as follows:\n", num_unv_succs);
		for (int i = 0; i < unvisited_succs.size(); i++) {
			msg("%i, ", unvisited_succs.at(i));
		}
		msg("\n");
	#endif
	if (num_unv_succs == 0) {
		#ifdef CDEBUG
			msg("none, so do nothing and return\n");
		#endif
		//no unvisited succs, do nothing
	}
	else if (num_unv_succs == 1) {
		//one unvisited successor, trivial recursion to it
		succnode = unvisited_succs[0];
		#ifdef CDEBUG
			msg("only %i, so visit this unvisited successor\n", unvisited_succs[0]);
		#endif
		if (accumulate_for_centroid_pdomseqnums_collect(blocks, postdoms, succnode, act_endofbranch, visited) == -1) return -1;
		
	}
	else { //more than one unvisited successor, determine branch order, recursively visit the branches in that order until
		//postdominator of branching reached and after the recursion proceed at that postdominator
		#ifdef CDEBUG
			msg("more than one unvisited successors, so calculate branch order\n");
		#endif
		int branching_postdom = postdoms[node];
		#ifdef CDEBUG
			msg("(postdom of that node is %i)\n", branching_postdom);
		#endif

		//determine branch order
		qvector<int> unvisited_succs_ordered;
		if(calculate_branchorder(blocks, &unvisited_succs, visited, branching_postdom, &unvisited_succs_ordered) == -1) return -1;
		#ifdef CDEBUG
			msg("calculated branch order as follows, i.e. first nodes of branches:\n");
			for (int i = 0; i < num_unv_succs; i++) {
				msg("%i , ", unvisited_succs_ordered[i]);
			}
			msg("\n");
		#endif
		//recursively visit branches in the calculated order, if they were not yet visited
		for (int i = 0; i < num_unv_succs; i++) {
			if (!visited[unvisited_succs_ordered[i]]) {
				#ifdef CDEBUG
					msg("successor branch beginning at %i still not visited, visit it now\n", unvisited_succs_ordered[i]);
				#endif
				if (accumulate_for_centroid_pdomseqnums_collect(blocks, postdoms, unvisited_succs_ordered[i], branching_postdom, visited) == -1) return -1;
				
			}
		}
		//proceed at branching postdominator, if it was not yet visited
		if (!visited[branching_postdom]) {
			#ifdef CDEBUG
				msg("all branches of node %i processed, so proceed at the nodes unvisited postdom %i\n", node, branching_postdom);
			#endif
			if (accumulate_for_centroid_pdomseqnums_collect(blocks, postdoms, branching_postdom, act_endofbranch, visited) == -1) return -1;
		}
		else {
			#ifdef CDEBUG
				msg("all branches of node %i processed, nodes postdom %i already visited\n", node, branching_postdom);
			#endif
			//do nothing
		}
	}

	return 0;
}

static void cen_analyze_naturalloop(qvector<qbasic_block_t>  *blocks, int node, int head_node, int* visited, bool accumulate) {

	//reverse DFS search back to loop head collects all loop nodes 
	qbasic_block_t *bb = &(blocks->at(node));
	visited[node] = 1;
	intseq_t npreds;
	intseq_t nsuccs;
	int prednode;
	int succnode;

	//check if node already part of loop with same head, if so these two loops are not nested => do not count them twice for node
	if (loopheads_by_node[node].count(head_node) == 0) {
		//no => new loop head found for that node
		loopheads_by_node[node].insert(head_node);
		if (accumulate) {
			//if accumulate param true, then for all edges the node is part of accumulate for centroids cz numerator
			npreds = bb->pred;
			for (int i = 0; i < npreds.size(); i++) {
				prednode = npreds[i];
				if (dfs_reachable_from_root[prednode]) {
					cznum_acc += numinstructs_by_node[node];
					cznum_dash_acc += (numinstructs_by_node[node] + numcalls_by_node[node]);
				}
			}
			nsuccs = bb->succ;
			for (int i = 0; i < nsuccs.size(); i++) {
				succnode = nsuccs[i];
				if (dfs_reachable_from_root[succnode]) {
					cznum_acc += numinstructs_by_node[node];
					cznum_dash_acc += (numinstructs_by_node[node] + numcalls_by_node[node]);
				}
			}
		}
		else {
			//if accumulate param false, only update loop depth info
			loopdepths_by_node[node]++;
		}
	}

	//recurse in direction to loop head
	if (node != head_node) {
		intseq_t npreds = bb->pred;
		int prednode;
		for (int i = 0; i < npreds.size(); i++) {
			prednode = npreds[i];
			if ((visited[prednode] == 0) && dfs_reachable_from_root[prednode]) { //only go to nodes reachable from root
				cen_analyze_naturalloop(blocks, prednode, head_node, visited, accumulate);
			}
		}
	}
}

static int initialize(qflow_chart_t *cfg, cfg_centroid_t* cen, cfg_centroid_t* cen_dash) {
	if ((cfg == NULL) || (cen == NULL) || (cen_dash == NULL)) {
		msg("failed to calculate centroid, bad function parameters\n");
		return -1;
	}
	cfg_num_nodes = cfg->size();
	cfg_blocks = &(cfg->blocks);
	if ((cfg_num_nodes <= 0) || (cfg_blocks->size() < cfg_num_nodes)) {
		msg("failed to retrieve the basic blocks of the function's cfg\n");
		return -1;
	}
	immdoms = new int[cfg_num_nodes + 1];
	postdoms = new int[cfg_num_nodes + 1];
	dfs_reachable_from_root = new bool[cfg_num_nodes + 1];
	revdfs_reachable_from_exit = new bool[cfg_num_nodes + 1];
	seqnums_by_node = new int[cfg_num_nodes + 1];
	numsuccs_by_node = new int[cfg_num_nodes + 1];
	for (int i = 0; i <= cfg_num_nodes; i++) {
		numsuccs_by_node[i] = 0;
	}
	numinstructs_by_node = new int[cfg_num_nodes + 1];
	numcalls_by_node = new int[cfg_num_nodes + 1];
	loopdepths_by_node = new int[cfg_num_nodes + 1];
	for (int i = 0; i <= cfg_num_nodes; i++) {
		loopdepths_by_node[i] = 0; 
	}
	loopheads_by_node = new std::set<int>[cfg_num_nodes + 1];
	label1 = new int[cfg_num_nodes + 1];
	label2 = new int[cfg_num_nodes + 1];
	for (int i = 0; i <= cfg_num_nodes; i++) {
		numinstructs_by_node[i] = -1; 
	}
	cxnum_acc = cynum_acc = cznum_acc = wbar_acc = 0;
	wbar_dash_acc = cxnum_dash_acc = cynum_dash_acc = cznum_dash_acc = 0;
	return 0;
}
static void cleanup() {
	delete[] label1;
	delete[] label2;
	delete[] seqnums_by_node;
	delete[] numsuccs_by_node;
	delete[] numinstructs_by_node;
	delete[] numcalls_by_node;
	delete[] loopdepths_by_node;
	delete[] loopheads_by_node;
	delete[] immdoms;
	delete[] postdoms;
	delete[] dfs_reachable_from_root;
	delete[] revdfs_reachable_from_exit;
}