#include "cfgtools.h"
#include <vector>
#include <set>

struct _cfg_branchorder_info {
	int br_startnode;
	int br_numnodes;
	int br_numstatements;
	unsigned __int64 br_1ststatement_binvalue;
};

typedef struct _cfg_branchorder_info cfg_branchorder_info_t;

struct compare_cfg_branchorder_infos {
	inline bool operator()(cfg_branchorder_info_t bri1, cfg_branchorder_info_t bri2)
	{
		if (bri1.br_numnodes != bri2.br_numnodes) {
			return bri1.br_numnodes < bri2.br_numnodes;
		}
		else {
			if (bri1.br_numstatements != bri2.br_numstatements) {
				return bri1.br_numstatements < bri2.br_numstatements;
			}
			else {
				if (bri1.br_1ststatement_binvalue != bri2.br_1ststatement_binvalue) {
					return bri1.br_1ststatement_binvalue < bri2.br_1ststatement_binvalue;
				}
				else {
					return true;
				}
			}
		}
	}
};

static void calc_postorder_and_reachability_dfs(qvector<qbasic_block_t>  *blocks, int node, bool *visited, int *num_reachable_nodes,
	int *nodes_by_postnum, int *postnums_by_node, int * postnum_counter);
static void calc_postorder_and_reachability_rev_dfs(qvector<qbasic_block_t>  *blocks, int node, bool *visited, int *num_reachable_nodes,
	int *nodes_by_postnum, int *postnums_by_node, int * postnum_counter,
	bool * dfs_reachable_from_root);
static int calc_branchorder_inspect_branch_dfs(qvector<qbasic_block_t> *blocks, int node,
	int *stopnodes, int branching_postdom, int *visited,
	int *acc_branch_numnodes, int *acc_branch_numstatements);

void analyze_dominance(qvector<qbasic_block_t>  *cfg_blocks, int *immdoms, bool* reachable_nodes, int *num_nodes_reachable) {
	int cfg_num_nodes = cfg_blocks->size();
	int postnum_counter;
	int local_num_nodes_reachable;
	int *immdoms_by_postnum;
	int *nodes_by_postnum;;
	int *postnums_by_node;
	//po_nums begin at 1, so following two arrays 1 slot larger
	immdoms_by_postnum = new int[cfg_num_nodes + 1];
	nodes_by_postnum = new int[cfg_num_nodes + 1];
	postnums_by_node = new int[cfg_num_nodes];

	//first do a dfs-search to calculate dfs-reachable nodes and their post order numbers 
	(*num_nodes_reachable) = 0;
	for (int i = 0; i < cfg_num_nodes; i++) {
		reachable_nodes[i] = false;
	}
	postnum_counter = 1; //dfs postorder counter, po nums start at 1
	calc_postorder_and_reachability_dfs(cfg_blocks, 0, reachable_nodes, num_nodes_reachable,
		nodes_by_postnum, postnums_by_node, &postnum_counter);
	local_num_nodes_reachable = (*num_nodes_reachable);
	
	#ifdef CDEBUG
		msg("%u nodes reachable, as follows:\n", local_num_nodes_reachable);
		for (int i = 0; i < cfg_num_nodes; i++) {
			msg("node %i: %d\n", i, reachable_nodes[i]);
		}
		msg("set postorder numbers:\n");
		for (int po_num = 1; po_num <= local_num_nodes_reachable; po_num++) {
			int node = nodes_by_postnum[po_num];
			msg("node %u with post order number %i\n", nodes_by_postnum[po_num], po_num);
		}
	#endif


	//initialize immdoms array index by postorder numbers, nums beginning with 1 for faster tests for zero 
	//all reachable and unreachable nodes initialize to 0 
	for (int i = 0; i < local_num_nodes_reachable; i++){
		immdoms_by_postnum[i] = 0;
	}
	//root node immdom of itself and has highest post-order number
	immdoms_by_postnum[local_num_nodes_reachable] = local_num_nodes_reachable;

	#ifdef CDEBUG
		msg("initial immdoms array for reachable nodes, indexed by post order numbers:\n");
		for (int po_num = 1; po_num <= local_num_nodes_reachable; po_num++) {
			msg("%i ", immdoms_by_postnum[po_num]);
		}
		msg("\n");
	#endif

	//something changed in last iteration ?
	bool changed;

	int act_node;
	qbasic_block_t *act_block;
	int new_immdom;
	int pred_ponum;
	int finger1, finger2;

	do {
		#ifdef CDEBUG
			msg("start dominance fixpoint iteration\n");
		#endif
		changed = false;
		// fetch all nodes in rev. post order, immdom entry of node is found by taking the intersection of its predecessors,
		// process is repeated until no changes were made 
		for (int act_ponum = local_num_nodes_reachable - 1; act_ponum >= 1; act_ponum--) {
			act_node = nodes_by_postnum[act_ponum];
			act_block = &(cfg_blocks->at(act_node));

			// first set immdom of actnode  = first processed reachable predecessor of actnode 
			//(at least one processed cause of its higher postorder number) 
			intseq_t preds = act_block->pred;
			int i = 0;
			int act_pred;
			for (; i < preds.size(); i++) {
				act_pred = preds[i];
				if (!reachable_nodes[act_pred]) continue;
				pred_ponum = postnums_by_node[act_pred];
				if (immdoms_by_postnum[pred_ponum] != 0) { //processed (in any iteration) pred ?
					new_immdom = pred_ponum;
					break;
				}
			}
			// then for all other already processed predecessors do an intersection
			for (; i < preds.size(); i++) {
				act_pred = preds[i];
				if (!reachable_nodes[act_pred]) continue;
				pred_ponum = postnums_by_node[act_pred];
				if (immdoms_by_postnum[pred_ponum] != 0) {//processed (in any iteration) pred ?
					// intersection: 
					// traverse currently set immdom of actnode and immdom of predecessor backwards to root until 
					// common immdom found, that's the resulting immdom of actnode
					finger1 = pred_ponum;
					finger2 = new_immdom;
					while (finger1 != finger2) {
						while (finger1 < finger2) finger1 = immdoms_by_postnum[finger1];
						while (finger2 < finger1) finger2 = immdoms_by_postnum[finger2];
					}
					new_immdom = finger1;
				}
			}
			//finally for that node check if the new immdom represents a change 
			if (immdoms_by_postnum[act_ponum] != new_immdom) {
				immdoms_by_postnum[act_ponum] = new_immdom;
				changed = true;
			}

		} //end for all nodes in one iteration
		#ifdef CDEBUG
			msg("dominance fixpoint iteration finished, new immdoms array for reachable nodes ind. by po numbers:\n");
			for (int po_num = 1; po_num <= local_num_nodes_reachable; po_num++) {
				msg("%i ", immdoms_by_postnum[po_num]);
			}
			msg("\n");
		#endif
	} while (changed);
	//end dom algo, now fill immdoms by node out param, unreachable nodes immdoms of themselfs
	for (int node = 0; node < cfg_num_nodes; node++) {
		if (reachable_nodes[node]) {
			immdoms[node] = nodes_by_postnum[immdoms_by_postnum[postnums_by_node[node]]];
		}
		else {
			immdoms[node] = node;
		}
	}
	#ifdef CDEBUG
		msg("calculated immdoms tree array ind. by nodes:\n");
		for (int node = 0; node < cfg_num_nodes; node++) {
			msg("%i ", immdoms[node]);
		}
		msg("\n");
	#endif

	delete[] immdoms_by_postnum;
	delete[] nodes_by_postnum;
	delete[] postnums_by_node;

}

//dfs post order number and reachability calculation
static void calc_postorder_and_reachability_dfs(qvector<qbasic_block_t>  *blocks, int node, bool *visited, int *num_reachable_nodes,
	int *nodes_by_postnum, int *postnums_by_node, int * postnum_counter){

	//mark as visited/update reachability info
	visited[node] = true;
	(*num_reachable_nodes) = (*num_reachable_nodes) + 1;

	//recurse
	qbasic_block_t *act_block = &(blocks->at(node));
	intseq_t succs = act_block->succ;
	int num_succs = succs.size();
	int succnode;
	for (int i = 0; i < num_succs; i++) {
		succnode = succs[i];
		if (visited[succnode] == 0) {
			calc_postorder_and_reachability_dfs(blocks, succnode, visited, num_reachable_nodes,
				nodes_by_postnum, postnums_by_node, postnum_counter);
		}
	}

	//set postorder number of node here, as all succs now processed
	nodes_by_postnum[(*postnum_counter)] = node;
	postnums_by_node[node] = (*postnum_counter);
	(*postnum_counter) = (*postnum_counter) + 1;
}

//counts the number of instructions and calls in a cfg basic block
int count_basicblock_instructions(qbasic_block_t* bb, int *num_instr, int *num_calls) {

	#ifdef CDEBUG
		msg("start counting instructions of basic block, address range %a - %a ...\n", bb->startEA, bb->endEA);
	#endif

	int num_bb_instrs = 0;
	int num_bb_calls = 0;


	/*no empty bbs should exist,
	traverse bb via start addresses of "items" (code or data, see bytes.hpp),
	endEA should be adress !after bb w.r.t.area.h ("excluded", however my tests found the following:
	. both for x86 and arm functions with startEA = endEA, so use "<= endEA" for loop test to avoid counting 0 instructions fpr bbs
	- for Android- ELF for ARM blocks with 1 instruction (instr. BX) with endEA = startEA + 4, so with "<=" wrongly instruction count 2
	=> TODO: how to code loop test (one possiblity use test_ea_end of SDK-API?, if wrong however: what's better for centroid calculation) ?*/
	for (ea_t addr = bb->startEA; addr <= bb->endEA;) {
		flags_t flags = getFlags(addr);
		if (flags == 0) {
			msg("failed to retrieve the flags of a basic block item's start address\n");
			return -1;
		}
		// check if we are at head byte of code, i.e. start byte of instruction 
		if (isHead(flags) && isCode(flags)) {
			num_bb_instrs++;
			//is instruction a call ?
			char mnem[MAXSTR];
			// fill the cmd structure with the disassembly of the current address and get the mnemonic text
			if (ua_mnem(addr, mnem, sizeof(mnem) - 1) == NULL) {
				msg("failed to decode the mnemonics of a basic block's instruction");
				return -1;
			}
			// check the mnemonic of the address against call mnemonics
			if (cmd.itype == NN_call || cmd.itype == NN_callfi || cmd.itype == NN_callni)
				num_bb_calls++;
		}
		addr = next_head(addr, 0xffffffff);
		//check w.r.t. IDAs address space boundary
		if (addr == BADADDR) break;
	}
	//set out parameters
	(*num_instr) = num_bb_instrs;
	(*num_calls) = num_bb_calls;
	return 0;
}

/*appends an exit node to the blocks of a cfg which are dfs-reachable from root*/
int append_exitnode(qflow_chart_t *cfg, bool *dfs_reachable_from_root) {
	//add exit node to all reachable nodes without successor or if they are ret blocks
	qvector<qbasic_block_t>  *cfg_blocks = &(cfg->blocks);
	qbasic_block_t exit_block;
	cfg_blocks->push_back(exit_block);
	int exit_block_index = cfg_blocks->size() - 1;
	intseq_t *exit_blocks_preds = &(cfg_blocks->at(exit_block_index).pred);
	for (int i = 0; i < exit_block_index; i++) {  
		if ((cfg_blocks->at(i).succ.size() == 0 || cfg->is_ret_block(i)) && dfs_reachable_from_root[i]) {
			#ifdef CDEBUG
				msg("block without successor or with ret statement found, append exit node\n");
			#endif
			cfg_blocks->at(i).succ.push_back(exit_block_index);
			//add block i as predecessor of exit block
			exit_blocks_preds->push_back(i);
		}
	}
	//exit node noq also reachable from node 0
	dfs_reachable_from_root[exit_block_index] = true;

	return 0;
}

int analyze_postdominance_from_exitnode(qvector<qbasic_block_t>  *cfg_blocks, bool * dfs_reachable_from_root, int *postdoms, 
	bool* revdfs_reachable_from_exit, int *num_nodes_revdfs_reachable_from_exit){
	
	int cfg_num_nodes = cfg_blocks->size();
	int postnum_counter;
	int local_num_nodes_revreachable;
	int *postdoms_by_postnum;
	int *nodes_by_postnum;
	int *postnums_by_node;
	//po_nums begin at 1, so following two arrays 1 slot larger
	postdoms_by_postnum = new int[cfg_num_nodes + 1];
	nodes_by_postnum = new int[cfg_num_nodes + 1];
	postnums_by_node = new int[cfg_num_nodes];

	//first do a dfs-search to calculate rev. dfs-reachable nodes and their post order numbers 
	(*num_nodes_revdfs_reachable_from_exit) = 0;
	for (int i = 0; i < cfg_num_nodes; i++) {
		revdfs_reachable_from_exit[i] = false;
	}
	postnum_counter = 1; //dfs postorder counter, po nums start at 1
	calc_postorder_and_reachability_rev_dfs(cfg_blocks, cfg_num_nodes-1, revdfs_reachable_from_exit, num_nodes_revdfs_reachable_from_exit,
		nodes_by_postnum, postnums_by_node, &postnum_counter, dfs_reachable_from_root);

	//check if all nodes dfs-reachable from root are also rev. dfs-reachable from exit, if not bad CFG -> return -1
	for (int i = 0; i < cfg_num_nodes; i++) {
		#ifdef CDEBUG
			msg("node %i dfs-reachable ?: %d\n", i, dfs_reachable_from_root[i]);
			msg("node %i: reachable from exit node ?: %d\n", i, revdfs_reachable_from_exit[i]);
		#endif
		if (dfs_reachable_from_root[i] != revdfs_reachable_from_exit[i]) {
			#ifdef CDEBUG
				msg("node %i not visible from both sides, so go along with DFS-seqnums", i);
			#endif
			delete[] postdoms_by_postnum;
			delete[] nodes_by_postnum;
			delete[] postnums_by_node;
			return -1;
		}
	}

	local_num_nodes_revreachable = (*num_nodes_revdfs_reachable_from_exit);

	#ifdef CDEBUG 
		msg("set rev. postorder numbers and initialized postdoms array:\n");
		for (int po_num = 1; po_num <= local_num_nodes_revreachable; po_num++) {
			msg("node %i got post order number %i\n", nodes_by_postnum[po_num], po_num);
		}
	#endif

	//initialize postdoms_array, nums beginning with 1 for faster tests for zero,
	for (int po_num = 1; po_num <= cfg_num_nodes; po_num++){
		postdoms_by_postnum[po_num] = 0;
	}
	//exit node postdom of itself and has highest post-order number
	postdoms_by_postnum[local_num_nodes_revreachable] = local_num_nodes_revreachable; 

	#ifdef CDEBUG
		msg("initial reachable nodes postdoms array indexed by post order numbers:\n");
		for (int po_num = 1; po_num <= cfg_num_nodes; po_num++) {
			msg("%i ", postdoms_by_postnum[po_num]);
		}
		msg("\n");
	#endif

		//something changed in last iteration ?
		bool changed;

		int act_node;
		qbasic_block_t *act_block;
		int new_postdom;
		int succ_ponum;
		int finger1, finger2;
		do {
			#ifdef CDEBUG
				msg("start postdominance fixpoint iteration\n");
			#endif
			changed = false;
			// fetch all nodes in reverse post order , postdom entry of a node is found by taking the intersection 
			// of its successors, the process is repeated until no changes are made 

			for (int act_ponum = local_num_nodes_revreachable - 1 /*skip exit node, never changes*/; act_ponum >= 1; act_ponum--) {
				act_node = nodes_by_postnum[act_ponum];
				act_block = &(cfg_blocks->at(act_node));
				// first set postdom of actnode  = first processed successor of actnode 
				//(at least one processed cause of its higher postorder num) 
				intseq_t succs = act_block->succ;
				int i = 0;
				for (; i < succs.size(); i++) {
					succ_ponum = postnums_by_node[succs[i]];
					if (postdoms_by_postnum[succ_ponum] != 0) { //processed (in any iteration) succ ?
						new_postdom = succ_ponum;
						break;
					}
				}
				// then for all other already processed successors do an intersection
				for (; i < succs.size(); i++) {
					succ_ponum = postnums_by_node[succs[i]];
					if (postdoms_by_postnum[succ_ponum] != 0) {//processed (in any iteration) succ ?
						// intersection: 
						// traverse both currently set postdom of actnode and postdom of predecessor backwards to exit 
						// node until common postdom found, that's the resulting postdom of actnode
						finger1 = succ_ponum;
						finger2 = new_postdom;
						while (finger1 != finger2) {
							while (finger1 < finger2) finger1 = postdoms_by_postnum[finger1];
							while (finger2 < finger1) finger2 = postdoms_by_postnum[finger2];
						}
						new_postdom = finger1;
					}
				}
				//finally for that node check if new postdom represents a change 
				if (postdoms_by_postnum[act_ponum] != new_postdom) {
					postdoms_by_postnum[act_ponum] = new_postdom;
					changed = true;
				}

			} //end for all nodes in one iteration
			#ifdef CDEBUG
				msg("postdominance fixpoint iteration finished, new reachable nodes postdoms array ind. by po numbers:\n");
				for (int po_num = 1; po_num <= local_num_nodes_revreachable; po_num++) {
					msg("%i ", postdoms_by_postnum[po_num]);
				}
				msg("\n");
			#endif
		} while (changed);
		//end postdom algo - postdoms tree array indexed by po nums calculated

		//fill a postdoms array indexed by node numbers 
		for (int node = 0; node < cfg_num_nodes; node++) {
			if (revdfs_reachable_from_exit[node]) {
				postdoms[node] = nodes_by_postnum[postdoms_by_postnum[postnums_by_node[node]]];
			}
			else {
				postdoms[node] = node; //set postdom of unreachable node to node
			}
		}

		#ifdef CDEBUG
			msg("calculated postdoms tree array ind. by nodes including exit node:\n");
			for (int node = 0; node < cfg_num_nodes; node++) {
				msg("%i ", postdoms[node]);
			}
			msg("\n");
		#endif

		delete[] postdoms_by_postnum;
		delete[] nodes_by_postnum;
		delete[] postnums_by_node;

		return 0;
}

//at begin of postdominance algorithm do a rev. dfs-search to calculate rev. post order numbers of nodes of the cfg
static void calc_postorder_and_reachability_rev_dfs(qvector<qbasic_block_t>  *blocks, int node, bool *visited, int *num_reachable_nodes,
	int *nodes_by_postnum, int *postnums_by_node, int * postnum_counter,
	bool * dfs_reachable_from_root) {

	//mark as visited
	visited[node] = true;
	qbasic_block_t *act_block = &(blocks->at(node));

	(*num_reachable_nodes) = (*num_reachable_nodes) + 1;

	//recurse
	intseq_t preds = act_block->pred;
	int num_preds = preds.size();
	int prednode;
	for (int i = 0; i < num_preds; i++) {
		prednode = preds[i];
		if (visited[prednode] == 0 && dfs_reachable_from_root[prednode]) { //only go to nodes reachable from node 0
				calc_postorder_and_reachability_rev_dfs(blocks, prednode, visited, num_reachable_nodes,
					nodes_by_postnum, postnums_by_node, postnum_counter, dfs_reachable_from_root);
		}
	}

	//set postorder number of node here, as all preds now processed
	nodes_by_postnum[(*postnum_counter)] = node;
	postnums_by_node[node] = (*postnum_counter);
	(*postnum_counter) = (*postnum_counter) + 1;
}

int calculate_branchorder(qvector<qbasic_block_t> *blocks, qvector<int> *branch_startnodes,
	int *processed_nodes, int branching_postdom, qvector<int> *branch_startnodes_ordered) {

	qvector<cfg_branchorder_info_t> branch_infos;
	cfg_branchorder_info_t  actbranch_info;
	int cfg_num_nodes = blocks->size();
	int *label = new int[cfg_num_nodes];
	int num_branches = branch_startnodes->size();
	int actbranch_startnode;
	int actbranch_numnodes;
	int actbranch_numstatements;
	unsigned __int64 actbranch_1ststatement_binvalue;
	ea_t act_branch_startblock_addr;
	ea_t act_branch_startblock_2ndin_addr;
	int instr_bytelength;
	char * byte_pointer;

	//for all branches ...
	for (int i = 0; i < num_branches; i++) {
		int actbranch_startnode = branch_startnodes->at(i);
		#ifdef CDEBUG
			msg("start doing the branch info calculation for branch starting at successor node %i\n", actbranch_startnode);
#		endif
		//... collect number of nodes (dfs-search stopping before nodes already given seqnums and branching postdom) , 
		//number of instructions in dfs and get binary values of first statements

		//calculate binary value of 1st statement: directly copy bytes of instruction into unsigned 64 bit,
		actbranch_1ststatement_binvalue = 0;
		qbasic_block_t *branch_startblock = &(blocks->at(actbranch_startnode));
		act_branch_startblock_addr = branch_startblock->startEA;
		int instr_size = decode_insn(act_branch_startblock_addr);
		byte_pointer = (char *)&actbranch_1ststatement_binvalue;
		//copy no more than 8 bytes
		if (instr_size > 8) {
			if (get_many_bytes(act_branch_startblock_addr, byte_pointer, 8) == 0) {
				msg("failed to fetch the bytes of an instruction during cfg branch ordering\n");
				delete[] label;
				return -1;
			}
		}
		else if (instr_size > 0){
			if (get_many_bytes(act_branch_startblock_addr, byte_pointer, 8) == 0) {
				msg("failed to fetch the bytes of an instruction during cfg branch ordering\n");
				delete[] label;
				return -1;
			}
			actbranch_1ststatement_binvalue >> (8 - instr_size);
		}
		else {
			actbranch_1ststatement_binvalue = 0;
		}

		actbranch_numnodes = 0;
		actbranch_numstatements = 0;

		//if the act branch begins at branching postdom, then numnodes and numstatements 0,so do no dfs
		if (actbranch_startnode != branching_postdom) {
			//dfs init
			for (int j = 0; j < cfg_num_nodes; j++) {
				label[j] = 0;
			}
			//do the dfs
			if (calc_branchorder_inspect_branch_dfs(blocks, actbranch_startnode,
				processed_nodes, branching_postdom, label,
				&actbranch_numnodes, &actbranch_numstatements) == -1) return -1;

		}
		else {
#ifdef CDEBUG
			msg("this successor is the postdom %i of the branching, so no recursive info collection\n", branching_postdom);
#endif
		}

		//store the info of the act branch for later branch comparison (entries ordered by entry order in branch_startnodes array)
		actbranch_info.br_startnode = actbranch_startnode; //node number
		actbranch_info.br_numnodes = actbranch_numnodes;
		actbranch_info.br_numstatements = actbranch_numstatements;
		actbranch_info.br_1ststatement_binvalue = actbranch_1ststatement_binvalue;
		branch_infos.push_back(actbranch_info);
#ifdef CDEBUG
		msg("branch info calculated for unvisited successor %i:\n", actbranch_startnode);
		msg("numnodes: %i, numstatements: %i, actbranch_1ststatement_binvalue: %llu\n", actbranch_numnodes,
			actbranch_numstatements, actbranch_1ststatement_binvalue);
#endif

	}
	//sort the branches with the collected infos and set out parameter (vector of node numbers)
	std::sort(branch_infos.begin(), branch_infos.end(), compare_cfg_branchorder_infos());
	for (int i = num_branches - 1; i >= 0; i--) { //branches with lower values have lower index after ordering
		branch_startnodes_ordered->push_back(branch_infos.at(i).br_startnode);
	}
	//done
	delete[] label;
	return 0;
}

//accumulates numbers of nodes and number of instructions of a cfg branch starting at "node" by dfs-crawling, stops before 
//"stopnodes" and the branching post dominator
static int calc_branchorder_inspect_branch_dfs(qvector<qbasic_block_t> *blocks, int node,
	int *stopnodes, int branching_postdom, int *visited,
	int *acc_branch_numnodes, int *acc_branch_numstatements) {

	if (node == blocks->size() - 1) return 0; //exit node should never be reached
	//mark node as visited
	visited[node] = 1;
	//accumulate
	qbasic_block_t *act_block = &(blocks->at(node));
	int actblock_numinstructs, actblock_numcalls;
	(*acc_branch_numnodes) = (*acc_branch_numnodes) + 1;
	if (count_basicblock_instructions(act_block, &(actblock_numinstructs), &(actblock_numcalls)) == -1) {
		msg("failed analyzing the instructions of a basic block\n");
		return -1;
	}
	(*acc_branch_numstatements) = (*acc_branch_numstatements) + actblock_numinstructs;
#ifdef CDEBUG
	msg("recursive branch info collection crawled to node %i, now %i branch num nodes and %i branch statements\n", node,
		(*acc_branch_numnodes), (*acc_branch_numstatements));
#endif
	//recurse, dont visit nodes already having seqnum (processed), postdom of branching and during this dfs already visited nodes
	intseq_t succs = act_block->succ;
	int num_succs = succs.size();
	int succnode;
	for (int i = 0; i < num_succs; i++) {
		succnode = succs[i];
		if ((stopnodes[succnode] == 0) && (succnode != branching_postdom) && (visited[succnode] == 0))
			if (calc_branchorder_inspect_branch_dfs(blocks, succnode,
				stopnodes, branching_postdom, visited,
				acc_branch_numnodes, acc_branch_numstatements) == 1) return -1;
	}
	return 0;
}